from ..types import Document, DocumentSet, DocumentIdentifier, Author
import requests
import shelve
import logging


class DBLPDocument(Document):
    def __init__(self, entry, authors):
        id = DocumentIdentifier(entry['title'], doi=entry.get('doi'))
        super().__init__(id)

        self.entry = entry
        self._authors = authors

    @property
    def title(self):
        return self.entry.get('title')

    @property
    def publication_year(self):
        try:
            return int(self.entry.get('year'))
        except Exception:
            return None

    @property
    def publication_type(self):
        return self.entry.get('type')

    @property
    def publication_source(self):
        return self.entry.get('venue')

    @property
    def publisher(self):
        return self.entry.get('publisher')

    @property
    def authors(self):
        return self._authors

    def __repr__(self):
        return f'<{self.title}>'


class DBLPAuthor(Author):
    def __init__(self, pid, name):
        self._pid = pid
        self._name = name

    @property
    def pid(self):
        return self._pid

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return f'<{self.name}>'


def process_authors(entry, author_cache):
    # Sometimes, authors is not a valid key
    if 'authors' not in entry:
        return None

    inputs = entry['authors']['author']
    outputs = []

    # Sometimes, inputs is nothing? (empty str or null)
    if not inputs:
        return None

    # Sometimes, inputs is str? (single author name)
    if isinstance(inputs, str):
        return [DBLPAuthor(None, inputs)]

    # Sometimes, input is dict? (single-item list)
    if isinstance(inputs, dict):
        inputs = [inputs]

    # Sometimes inputs is list
    for author in inputs:
        pid = author['@pid']
        name = author['text']

        if pid not in author_cache:
            author_cache[pid] = DBLPAuthor(pid, name)

        outputs.append(author_cache[pid])

    return outputs


CACHE_FILE = '.dblp'
DBLP_URL = 'http://dblp.org/search/publ/api'


def search_dblp(query: str, *, limit=None) -> DocumentSet:
    """Perform the given `query` on the DBLP API and return the results
    as a `DocumentSet`.

    :param limit: The maximum number of documents to retrieve.
    """

    attr = dict(format='json', h=100, q=query, f=0)
    offset = 0

    docs = []
    author_cache = dict()

    with shelve.open(CACHE_FILE) as cache:
        while True:
            key = f'{query};{offset}'

            if key not in cache:
                attr['f'] = offset
                req = requests.get(DBLP_URL, params=attr)
                cache[key] = req.json().get('result')

            data = cache[key]

            if not data:
                break

            status = data.get('status').get('text')
            if status != 'OK':
                logging.warning(f'expecting status OK, got status {status}')
                break

            if 'hits' not in data or 'hit' not in data['hits']:
                break

            entries = data['hits']['hit']
            offset += len(entries)

            if not entries:
                break

            for entry in entries:
                entry = entry['info']

                authors = process_authors(entry, author_cache)
                docs.append(DBLPDocument(entry, authors))

            if limit is not None and len(docs) >= limit:
                break

    return DocumentSet(docs)



COUNTRY_TO_CONTINENT = {
        'afghanistan': 'Asia',
        'albania': 'Europe',
        'algeria': 'Africa',
        'american samoa': 'Oceania',
        'andorra': 'Europe',
        'angola': 'Africa',
        'anguilla': 'North America',
        'antigua and barbuda': 'North America',
        'argentina': 'South America',
        'armenia': 'Asia',
        'aruba': 'North America',
        'australia': 'Oceania',
        'austria': 'Europe',
        'azerbaijan': 'Asia',
        'bahamas': 'North America',
        'bahrain': 'Asia',
        'bangladesh': 'Asia',
        'barbados': 'North America',
        'belarus': 'Europe',
        'belgium': 'Europe',
        'belize': 'North America',
        'benin': 'Africa',
        'bermuda': 'North America',
        'bhutan': 'Asia',
        'bolivia, plurinational state of': 'South America',
        'bolivia': 'South America',
        'bonaire, sint eustatius and saba': 'North America',
        'bonaire': 'North America',
        'sint eustatius': 'North America',
        'saba': 'North America',
        'bosnia and herzegovina': 'Europe',
        'botswana': 'Africa',
        'brazil': 'South America',
        'british indian ocean territory': 'Asia',
        'brunei darussalam': 'Asia',
        'bulgaria': 'Europe',
        'burkina faso': 'Africa',
        'burundi': 'Africa',
        'cabo verde': 'Africa',
        'cambodia': 'Asia',
        'cameroon': 'Africa',
        'canada': 'North America',
        'cayman islands': 'North America',
        'cayman island': 'North America',
        'central african republic': 'Africa',
        'chad': 'Africa',
        'chile': 'South America',
        'china': 'Asia',
        'christmas island': 'Asia',
        'cocos (keeling) islands': 'Asia',
        'cocos keeling islands': 'Asia',
        'cocos islands': 'Asia',
        'colombia': 'South America',
        'comoros': 'Africa',
        'congo': 'Africa',
        'cook islands': 'Oceania',
        'costa rica': 'North America',
        'croatia': 'Europe',
        'cuba': 'North America',
        'curaçao': 'North America',
        'curacao': 'North America',
        'cyprus': 'Asia',
        'czechia': 'Europe',
        "côte d'ivoire": 'Africa',
        'denmark': 'Europe',
        'djibouti': 'Africa',
        'dominica': 'North America',
        'dominican republic': 'North America',
        'ecuador': 'South America',
        'egypt': 'Africa',
        'el salvador': 'North America',
        'equatorial guinea': 'Africa',
        'eritrea': 'Africa',
        'estonia': 'Europe',
        'eswatini': 'Africa',
        'ethiopia': 'Africa',
        'falkland islands (malvinas)': 'South America',
        'faroe islands': 'Europe',
        'fiji': 'Oceania',
        'finland': 'Europe',
        'france': 'Europe',
        'french guiana': 'South America',
        'french polynesia': 'Oceania',
        'gabon': 'Africa',
        'gambia': 'Africa',
        'georgia': 'Asia',
        'germany': 'Europe',
        'ghana': 'Africa',
        'gibraltar': 'Europe',
        'greece': 'Europe',
        'greenland': 'North America',
        'grenada': 'North America',
        'guadeloupe': 'North America',
        'guam': 'Oceania',
        'guatemala': 'North America',
        'guernsey': 'Europe',
        'guinea': 'Africa',
        'guinea-bissau': 'Africa',
        'guyana': 'South America',
        'haiti': 'North America',
        'honduras': 'North America',
        'hong kong': 'Asia',
        'hungary': 'Europe',
        'iceland': 'Europe',
        'india': 'Asia',
        'indonesia': 'Asia',
        'iran, islamic republic of': 'Asia',
        'iran': 'Asia',
        'repulic of iran': 'Asia',
        'iraq': 'Asia',
        'ireland': 'Europe',
        'isle of man': 'Europe',
        'israel': 'Asia',
        'italy': 'Europe',
        'jamaica': 'North America',
        'japan': 'Asia',
        'jersey': 'Europe',
        'jordan': 'Asia',
        'kazakhstan': 'Asia',
        'kenya': 'Africa',
        'kiribati': 'Oceania',
        "korea, democratic people's republic of": 'Asia',
        "democratic people's republic of korea": 'Asia',
        'korea, republic of': 'Asia',
        'korea': 'Asia',
        'republic of korea': 'Asia',
        'kuwait': 'Asia',
        'kyrgyzstan': 'Asia',
        "lao people's democratic republic": 'Asia',
        'latvia': 'Europe',
        'lebanon': 'Asia',
        'lesotho': 'Africa',
        'liberia': 'Africa',
        'libya': 'Africa',
        'liechtenstein': 'Europe',
        'lithuania': 'Europe',
        'luxembourg': 'Europe',
        'macao': 'Asia',
        'madagascar': 'Africa',
        'malawi': 'Africa',
        'malaysia': 'Asia',
        'maldives': 'Asia',
        'mali': 'Africa',
        'malta': 'Europe',
        'marshall islands': 'Oceania',
        'martinique': 'North America',
        'mauritania': 'Africa',
        'mauritius': 'Africa',
        'mayotte': 'Africa',
        'mexico': 'North America',
        'micronesia, federated states of': 'Oceania',
        'micronesia': 'Oceania',
        'moldova, republic of': 'Europe',
        'moldova': 'Europe',
        'monaco': 'Europe',
        'mongolia': 'Asia',
        'montenegro': 'Europe',
        'montserrat': 'North America',
        'morocco': 'Africa',
        'mozambique': 'Africa',
        'myanmar': 'Asia',
        'namibia': 'Africa',
        'nauru': 'Oceania',
        'nepal': 'Asia',
        'netherlands': 'Europe',
        'new caledonia': 'Oceania',
        'new zealand': 'Oceania',
        'nicaragua': 'North America',
        'niger': 'Africa',
        'nigeria': 'Africa',
        'niue': 'Oceania',
        'norfolk island': 'Oceania',
        'north macedonia': 'Europe',
        'northern mariana islands': 'Oceania',
        'norway': 'Europe',
        'oman': 'Asia',
        'pakistan': 'Asia',
        'palau': 'Oceania',
        'palestine, state of': 'Asia',
        'palestine': 'Asia',
        'panama': 'North America',
        'papua new guinea': 'Oceania',
        'paraguay': 'South America',
        'peru': 'South America',
        'philippines': 'Asia',
        'poland': 'Europe',
        'portugal': 'Europe',
        'puerto rico': 'North America',
        'qatar': 'Asia',
        'romania': 'Europe',
        'russian federation': 'Europe',
        'rwanda': 'Africa',
        'réunion': 'Africa',
        'saint barthélemy': 'North America',
        'saint helena, ascension and tristan da cunha': 'Africa',
        'saint kitts and nevis': 'North America',
        'saint lucia': 'North America',
        'saint martin (french part)': 'North America',
        'saint pierre and miquelon': 'North America',
        'saint vincent and the grenadines': 'North America',
        'samoa': 'Oceania',
        'san marino': 'Europe',
        'sao tome and principe': 'Africa',
        'saudi arabia': 'Asia',
        'senegal': 'Africa',
        'serbia': 'Europe',
        'seychelles': 'Africa',
        'sierra leone': 'Africa',
        'singapore': 'Asia',
        'slovakia': 'Europe',
        'slovenia': 'Europe',
        'solomon islands': 'Oceania',
        'somalia': 'Africa',
        'south africa': 'Africa',
        'south georgia and the south sandwich islands': 'South America',
        'south sudan': 'Africa',
        'spain': 'Europe',
        'sri lanka': 'Asia',
        'sudan': 'Africa',
        'suriname': 'South America',
        'svalbard and jan mayen': 'Europe',
        'sweden': 'Europe',
        'switzerland': 'Europe',
        'syrian arab republic': 'Asia',
        'taiwan, province of china': 'Asia',
        'taiwan': 'Asia',
        'tajikistan': 'Asia',
        'tanzania, united republic of': 'Africa',
        'tanzania': 'Africa',
        'thailand': 'Asia',
        'togo': 'Africa',
        'tokelau': 'Oceania',
        'tonga': 'Oceania',
        'trinidad and tobago': 'North America',
        'tunisia': 'Africa',
        'turkey': 'Asia',
        'turkmenistan': 'Asia',
        'turks and caicos islands': 'North America',
        'tuvalu': 'Oceania',
        'uganda': 'Africa',
        'ukraine': 'Europe',
        'united arab emirates': 'Asia',
        'united kingdom': 'Europe',
        'united states': 'North America',
        'uruguay': 'South America',
        'uzbekistan': 'Asia',
        'vanuatu': 'Oceania',
        'venezuela, bolivarian republic of': 'South America',
        'venezuela': 'South America',
        'viet nam': 'Asia',
        'vietnam': 'Asia',
        'virgin islands, british': 'North America',
        'virgin islands, u.s.': 'North America',
        'wallis and futuna': 'Oceania',
        'yemen': 'Asia',
        'zambia': 'Africa',
        'zimbabwe': 'Africa',
        'åland islands': 'Europe',
        'aland islands': 'Europe'
 }



from .types import DocumentMapping, DocumentSet
from collections import defaultdict
from matplotlib.colors import to_hex
import seaborn as sns
import json
import numpy as np
import networkx as nx
import pyvis
import textwrap


def calculate_layout(g, iterations=1000, gravity=1):
    try:
        from fa2 import ForceAtlas2
        model = ForceAtlas2(
                verbose=True,
                scalingRatio=1,
                gravity=gravity,
        )

        matrix = nx.to_scipy_sparse_matrix(g, dtype='f', format='lil',
                                           weight='weight')
        pos = model.forceatlas2(matrix, iterations=iterations)

        return dict(zip(g.nodes(), pos))

    except ImportError:
        return nx.drawing.layout.spring_layout(g)


def plot_network(g: nx.Graph, *, height='1000px', smooth_edges=None,
                 max_node_size=100, min_node_size=5, largest_component=True,
                 interactive=True, controls=False, scale=1, iterations=1000,
                 gravity=1):
    """Plot a `networkx.Graph` generated by one of the `build_X_graph`
    functions in this module. Plotting is done using the `pyvis` library.


    :param height: Height of the plot.
    :param smooth_edges: Enables curved ('smooth') edges. Looks nice but is
                         heavy on performance.
    :param min_node_size: The radius of the smallest node.
    :param max_node_size: The radius of the largest node.
    :param largest_component: Only plot the largest connected component of
                              the graph.
    """
    if isolates := list(nx.isolates(g)):
        g = g.copy()
        g.remove_nodes_from(isolates)

    if len(g.edges) == 0:
        print('no edges given')
        return

    directed = nx.is_directed(g)

    if largest_component:
        if directed:
            cc = nx.weakly_connected_components(g)
        else:
            cc = nx.connected_components(g)

        cc = sorted(cc, key=len, reverse=True)
        g = g.subgraph(cc[0])

    v = pyvis.network.Network(
            notebook=True,
            width='100%',
            height=height,
            directed=directed
    )

    sizes = [attr.get('size') or attr.get('weight') for (_, attr) in
             g.nodes.items()]

    if all(s is not None for s in sizes):
        sizes = np.array(sizes)
    elif directed:
        sizes = [g for (_, g) in g.in_degree]
    else:
        sizes = [g for (_, g) in g.degree]

    layout = calculate_layout(g, gravity=gravity, iterations=iterations)
    sizes = np.array(sizes, dtype=np.float32)
    ratio = (max_node_size - min_node_size) / np.amax(sizes)
    sizes = ratio * sizes + min_node_size

    for id, size in zip(g, sizes):
        attr = g.nodes[id]
        kwargs = dict(labelHighlightBold=True)

        kwargs['shape'] = attr.get('shape', 'dot')
        kwargs['title'] = attr['title']
        kwargs['label'] = textwrap.fill(attr['title'], width=20)
        kwargs['size'] = float(size)

        if layout is not None:
            pos = layout[id]
            kwargs['x'] = pos[0] * scale
            kwargs['y'] = pos[1] * scale

        color = attr.get('color')
        if color is not None:
            kwargs['color'] = to_hex(color)

        v.add_node(id, **kwargs)

    for src, dst in g.edges():
        weight = g[src][dst].get('weight')
        if weight is not None:
            width = weight
            label = str(weight)
        else:
            width = None
            label = ''

        v.add_edge(src, dst, width=width, title=label)

    if smooth_edges is None:
        smooth_edges = len(g.edges()) < 1000

    v.set_options(json.dumps({
        'configure': {
            'enabled': controls,
        },
        'nodes': {
            'font': {
                'size': 7,
            },
        },
        'edges': {
            'smooth': smooth_edges,
            'color': {
                'opacity': 0.25,
            }
        },
        'physics': {
            'enabled': interactive,
            'forceAtlas2Based': {
                'springLength': 100,
            },
            'solver': 'forceAtlas2Based',
        }
    }))

    return v.show('citation.html')


def build_base_network(docs, directed, colors=None, cmap=None,
                       shapes=None, sizes=None, node_props=None):
    def extract_column(docs, expr):
        if not docs:
            result = []
        elif not isinstance(expr, str):
            result = list(expr)
        elif expr in docs.data:
            result = docs.data[expr]
        else:
            result = docs.data.eval(expr)

        assert len(result) == len(docs)
        return result

    g = nx.DiGraph() if directed else nx.Graph()
    mapping = DocumentMapping()

    if node_props is None:
        node_props = docs.data.columns

    for i, doc in enumerate(docs):
        attr = dict()

        for prop in node_props:
            attr[prop] = docs.data[prop][i]

        g.add_node(i, title=doc.title, doc=doc, **attr)
        mapping.add(doc.id, i)

    # Custom colors
    if colors is not None:
        colors = extract_column(docs, colors)

        if all(isinstance(c, float) for c in colors):
            begin, end = min(colors), max(colors)
            cmap = sns.color_palette(cmap, as_cmap=True)
            colors = [cmap(float(c - begin) / (end - begin)) for c in colors]
        # if all(isinstance(c, int) for c in colors):
        else:
            groups = dict((c, i) for i, c in enumerate(sorted(set(colors))))
            cmap = sns.color_palette(cmap, n_colors=len(groups))
            colors = [cmap[groups[c]] for c in colors]

        for i, c in enumerate(colors):
            g.nodes[i]['color'] = c

    # Custom sizes
    if sizes is not None:
        sizes = extract_column(docs, sizes)

        for i, size in enumerate(sizes):
            g.nodes[i]['size'] = size

    if shapes is not None:
        shapes = extract_column(docs, shapes)

        for i, shape in enumerate(sizes):
            g.nodes[i]['shape'] = shape

    return g, mapping


def split_kwargs(*names, **kwargs):
    names = list(names) + ['colors', 'cmap']
    left, right = dict(), dict()

    for k, v in kwargs.items():
        if k in names:
            left[k] = v
        else:
            right[k] = v

    return left, right


def build_citation_network(docs: DocumentSet, **kwargs) -> nx.Graph:
    """Builds a citation network: a directed graph where each node
    corresponds to a document and each directed edge indicates that
    one document cites the other."""
    g, mapping = build_base_network(docs, True, **kwargs)

    for i, doc in enumerate(docs):
        for ref in doc.references or []:
            j = mapping.get(ref)

            if j is not None:
                g.add_edge(i, j)

    return g


def plot_citation_network(docs: DocumentSet, **kwargs):
    """Plot a citation network.

    This is a shorthand for `plot_network(build_citation_network(docs))`."""
    b, p = split_kwargs(**kwargs)
    return plot_network(build_citation_network(docs, **b), **p)


def build_cocitation_network(docs: DocumentSet, *, max_edges=None, **kwargs
                             ) -> nx.Graph:
    """Builds a co-citation network: a undirected graph where each node
    corresponds to a document and the edge weights stores the cocitation
    strengths (i.e., the number of times two documents have been cited
    together).

    :param max_edges: Select only the top edges. This is useful since
        cocitation networks are often dense and only the strongest edges
        are usually important.
    """
    max_edges = max_edges or len(docs) * 2

    g, mapping = build_base_network(docs, False, **kwargs)
    strength = defaultdict(int)

    for doc in docs:
        refs = []

        for ref in doc.references or []:
            j = mapping.get(ref)

            if j is not None:
                refs.append(j)

        for i in refs:
            for j in refs:
                if i < j:
                    strength[i, j] += 1

    edges = list(strength.items())

    if len(edges) > max_edges:
        edges.sort(key=lambda p: p[1], reverse=True)
        edges = edges[:max_edges]

    for (i, j), weight in edges:
        g.add_edge(i, j, weight=weight)

    return g


def plot_cocitation_network(docs: DocumentSet, *, max_edges=None,
                            node_size=10, **kwargs):
    """Plot a citation network.

    This is a shorthand for `plot_network(build_cocitation_network(docs))`."""
    b, p = split_kwargs(**kwargs)
    return plot_network(
            build_cocitation_network(docs, max_edges=max_edges, **b),
            # min_node_size=node_size,
            # max_node_size=node_size,
            **p
    )


def build_coupling_network(docs: DocumentSet, max_edges=1000, **kwargs
                           ) -> nx.Graph:
    """Builds a bibligraphic coupling network: an undirected graph where
    nodes indicate documents and edge weights store the bibliographic
    coupling strength. This strength measures how similar two documents
    view related work. It is measured as the number of shared references
    between two documents.

    :param max_edges: Select only the top edges. This is useful since these
        networks are often dense and only the strongest edges are usually
        important.
    """

    g, mapping = build_base_network(docs, False, **kwargs)
    n = len(g)
    doc_refs = []

    for doc in docs:
        refs = []

        for ref in doc.references or []:
            i = mapping.get(ref)

            if i is None:
                mapping.add(ref, n)
                n += 1
                i = n

            if i is not None:
                refs.append(i)

        doc_refs.append(set(refs))

    strength = defaultdict(int)

    for i, a in enumerate(doc_refs):
        for j, b in enumerate(doc_refs[:i]):
            common = a & b

            if common:
                strength[i, j] = len(common)

    edges = list(strength.items())

    if len(edges) > max_edges:
        edges.sort(key=lambda p: p[1], reverse=True)
        edges = edges[:max_edges]

    for (i, j), weight in edges:
        g.add_edge(i, j, weight=weight, score=weight)

    return g


def plot_coupling_network(docs: DocumentSet, *, max_edges=None, node_size=10,
                          **kwargs):
    """Plot a bibliographic coupling network.

    This is a shorthand for `plot_network(build_coupling_network(docs))`."""
    b, p = split_kwargs(**kwargs)
    return plot_network(
            build_coupling_network(docs, max_edges, **b),
            min_node_size=node_size,
            max_node_size=node_size,
            **p
    )


def build_coauthor_network(docs: DocumentSet, *, max_authors=None) -> nx.Graph:
    """Builds a co-author network: an undirected graph where nodes indicate
    authors and edge weight indicate the number of shared publications
    between two authors.

    :param max_authors: Select only the top X authors.
    """
    g = nx.DiGraph()
    count = defaultdict(int)

    for doc in docs:
        authors = []

        for author in doc.authors or []:
            name = author.name

            if name:
                count[name] += 1

    authors = list(count.keys())

    if max_authors is not None and len(authors) > max_authors:
        authors.sort(key=lambda name: count[name], reverse=True)
        authors = authors[:max_authors]

    mapping = dict()
    for i, author in enumerate(authors):
        g.add_node(i, title=author)
        mapping[author] = i

    edges = defaultdict(int)

    for doc in docs:
        authors = [a.name for a in doc.authors or [] if a.name]

        for i, left in enumerate(authors):
            for right in authors[:i]:
                if left in mapping and right in mapping:
                    edges[mapping[left], mapping[right]] += 1

    for (left, right), weight in edges.items():
        g.add_edge(left, right, weight=weight)

    return g


def plot_coauthor_network(docs: DocumentSet, *, max_authors=None, **kwargs):
    """Plot a co-author network.

    This is a shorthand for `plot_network(build_coauthor_network(docs))`."""
    b, p = split_kwargs(**kwargs)
    return plot_network(
            build_coauthor_network(docs, max_authors=max_authors, **b), **p
    )



from .stats import \
        compute_year_histogram, \
        compute_author_histogram, \
        compute_author_affiliation_histogram, \
        compute_language_histogram, \
        compute_number_authors_histogram, \
        compute_source_histogram, \
        compute_source_type_histogram, \
        compute_affiliation_histogram, \
        compute_country_histogram, \
        compute_continent_histogram, \
        compute_groups_histogram
from .nlp import \
        generate_topic_cloud, \
        calculate_embedding, \
        compute_word_distribution, \
        TopicModel, \
        Corpus
from .types import DocumentSet
import inspect
import math
import matplotlib.pyplot as plt
import numpy as np
import seaborn
import pandas as pd


def plot_histogram(
                   data: pd.DataFrame,
                   keys=None,
                   title='',
                   xlabel='',
                   ylabel='',
                   label_rotation=None,
                   vertical=False,
                   bar_width=0.8,
                   max_label_length=100,
                   stacked=False,
                   legend=True,
                   relative_to=None,
                   ax=None):
    """
    This is the general function to plot a histogram (bar plot). All other
    `plot_*_histogram` functions in this module will call this function.

    This function takes a pandas `DataFrame`. Each column represents one
    group and a sequence of bars that will be plotted. The names in the index
    are placed as labels on the axis.

    For instance, a possible input could be
    data frame where the columns are different authors, rows are different
    years, and values are the number of publications per year per author.

    :param ax: The `matplotlib` `Axes` instance the plot will be drawn on. If
               `None`, the current `Axes` instance is used (`plt.gca()`).
    :param vertical: Default bars are horizontal (left to right). Set
                     `vertical=True` for vertical bars (bottom to top).
    :param bar_width: Width of bars. Should be at most `1.0` for 100%.
    :param label_rotation: Rotates the xlabels. This is useful if
                           `vertical=True` since it can be used to place the
                           labels horziontal (`label_rotation=0`), vertical
                           (`label_rotation=90`), or diagonal
                           (`label_rotation=45`).
    :param max_label_length: Labels longer than this length are shortened.
    :param stacked: By default, different groups are drawn next to each other.
                    If `True`, the different groups are stacked on top of each
                    other instead.
    :param legend: Show legend.
    :param relative_to: If not `None`, all bars will be plotted as a
                        percentage relative to this value.
    :param title: Title of plot (set using `ax.set_title`).
    :param xlabel: Title on the X axis (or Y axis if `vertical=True`).
    :param ylabel: Title on the Y axis (or X axis if `vertical=True`).
    """
    if ax is None:
        ax = plt.gca()

    if label_rotation is None:
        if vertical:
            label_rotation = 90
        else:
            label_rotation = 0

    if not ylabel:
        if relative_to is not None:
            ylabel = '% of documents'
        else:
            ylabel = 'No. of documents'

    if not vertical:
        ax.grid(b=False, which='both', axis='y')
        xlabel, ylabel = ylabel, xlabel
    else:
        ax.grid(b=False, which='both', axis='x')

    if title:
        ax.set_title(title)

    if xlabel:
        ax.set_xlabel(xlabel)

    if ylabel:
        ax.set_ylabel(ylabel)

    if keys is None:
        keys = list(data.index)

    n = len(keys)
    groups = list(data)

    if not groups or not keys:
        return

    bar_bottom = np.array([0] * len(keys))

    for i, group in enumerate(groups):
        ys = data[group]
        bottom = bar_bottom

        if relative_to is not None:
            ys = ys / relative_to * 100

        if stacked:
            bar_bottom = bar_bottom + ys
            xs = np.arange(n) - 0.5 * bar_width
            bin_width = bar_width
        else:
            bin_width = bar_width / len(groups)
            xs = np.arange(n) + i * bin_width - 0.5 * bar_width

        if not vertical:
            ax.barh(xs, ys, label=group, height=bin_width, left=bottom,
                    align='edge')
        else:
            ax.bar(xs, ys, label=group, width=bin_width, bottom=bottom,
                   align='edge')

    if legend and len(groups) > 1:
        ax.legend()

    def shorten(label):
        label = str(label)
        n = max_label_length
        return label[:n] + '...' if len(label) > n else label

    labels = list(map(shorten, keys))

    if not vertical:
        ax.set_ylim((n - 0.5, -0.5))
        ax.set_yticks(range(n))
        ax.set_yticklabels(labels, rotation=label_rotation)
    else:
        ax.set_xlim((-0.5, n - 0.5))
        ax.set_xticks(range(n))
        ax.set_xticklabels(labels, rotation=label_rotation, ha='right')

    return ax


def wrapper(docs, fun, default, **kwargs):
    for key, value in default.items():
        kwargs.setdefault(key, value)

    plot_kwargs = dict()
    params = inspect.signature(plot_histogram).parameters

    for key in list(kwargs):
        if key in params:
            plot_kwargs[key] = kwargs.pop(key)

    if kwargs.pop('relative', False):
        plot_kwargs['relative_to'] = len(docs)

    data = fun(docs, **kwargs)
    return plot_histogram(data, **plot_kwargs)


def plot_groups_histogram(docs, **kwargs):
    """ """

    def fun(docs, **kwargs):
        return compute_groups_histogram(docs, **kwargs).T

    default = dict(title='Categories')
    return wrapper(docs, fun, default, **kwargs)


def plot_year_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents published in each year.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Year of publications')
    return wrapper(docs, compute_year_histogram, default, **kwargs)


def plot_author_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents published per author.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Authors', limit=25)
    return wrapper(docs, compute_author_histogram, default, **kwargs)


def plot_number_authors_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of authors per document.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='No. of authors')
    return wrapper(docs, compute_number_authors_histogram, default, **kwargs)


def plot_author_affiliation_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents published per author
    affiliation.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Author + Affiliation', limit=25)
    return wrapper(docs, compute_author_affiliation_histogram, default,
                   **kwargs)


def plot_language_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by language. """
    default = dict(title='Language')
    return wrapper(docs, compute_language_histogram, default, **kwargs)


def plot_source_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by publication source. """
    default = dict(title='Publication source', limit=25)
    return wrapper(docs, compute_source_histogram, default, **kwargs)


def plot_source_type_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by publication source
    type.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Publication source type', limit=25)
    return wrapper(docs, compute_source_type_histogram, default, **kwargs)


def plot_affiliation_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by author affiliation.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Affiliations', limit=25)
    return wrapper(docs, compute_affiliation_histogram, default, **kwargs)


def plot_country_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by country of author
    affiliation

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Countries', limit=25)
    return wrapper(docs, compute_country_histogram, default, **kwargs)


def plot_continent_histogram(docs: DocumentSet, **kwargs):
    """ Plot histogram of the number of documents by continent of author
    affiliation.

    :param kwargs: Passed to `plot_histogram`.
    """
    default = dict(title='Continents', limit=25)
    return wrapper(docs, compute_continent_histogram, default, **kwargs)


def plot_word_distribution(corpus: Corpus, *, limit=25, **kwargs):
    """ Plot the frequency of the top words in the given corpus.

    :param kwargs: Passed to `plot_histogram`.
    """
    n = len(corpus.frequencies)
    data = compute_word_distribution(corpus, limit=limit)
    return plot_histogram(data, relative_to=n, **kwargs)


def plot_embedding(corpus: Corpus, model: TopicModel, layout=None, ax=None):
    """ TODO """
    if ax is None:
        ax = plt.gca()

    if layout is None:
        layout = calculate_embedding(corpus)

    num_topics = len(model.topic2token)
    best_topic = np.argmax(model.doc2topic.T, axis=0)

    colors = seaborn.color_palette('hls', num_topics)
    colors = np.array(colors)[:, :3] * 0.9  # Mute colors a bit

    for i in range(num_topics):
        indices = best_topic == i
        # label = 'ABCDEFGHIJLMNOPQRSTUVWXYZ'[i]
        label = i + 1

        for j in np.argwhere(indices)[:, 0]:
            x, y = layout[j]
            ax.scatter(
                    x,
                    y,
                    marker='o',
                    s=150,
                    linewidth=0.5,
                    color=colors[i],
                    zorder=2*j,
            )

            ax.text(
                x,
                y,
                label,
                fontsize=6,
                color='1',
                va='center',
                ha='center',
                fontweight='bold',
                zorder=2*j + 1,
            )

        top_items = model.best_tokens_for_topic(i, limit=3)
        label = f'Topic {label}:' + ', '.join(top_items)

        center = np.median(layout[indices], axis=0)
        ax.text(
                center[0],
                center[1],
                label,
                va='center',
                ha='center',
                color='1',
                backgroundcolor=(0, 0, 0, .75),
                zorder=10 * len(best_topic),
        )

    ax.set_aspect('equal')
    ax.set_xticks([])
    ax.set_yticks([])


def plot_topic_clouds(model: TopicModel, *, fig=None, ncols=3, **kwargs):
    """ Plot word clouds for each of the topics from the given topic model. """
    if fig is None:
        plt.clf()
        fig = plt.gcf()

    nrows = math.ceil(model.num_topics / float(ncols))

    for i in range(model.num_topics):
        ax = fig.add_subplot(nrows, ncols, i + 1)
        plot_topic_cloud(model, i, ax=ax, **kwargs)


def plot_topic_cloud(model: TopicModel, topic_id: int, *, ax=None, **kwargs):
    """ Plot a word cloud for the given topic from the given topic model. """
    if ax is None:
        ax = plt.gca()

    ax.set_xticks([])
    ax.set_yticks([])

    im = generate_topic_cloud(model, topic_id, **kwargs).to_array()
    ax.set_title(f'Topic {topic_id + 1}')
    ax.imshow(im, interpolation='bilinear')


def plot_document_topics(model: TopicModel, document_id: int, *, ax=None):
    if ax is None:
        ax = plt.gca()

    weights = model.document_topics(document_id)

    ax.bar(
            np.arange(model.num_topics),
            model.document_topics(document_id)
    )
    ax.set_ylim(0, 1)
    ax.set_ylabel('Relevance')
    ax.set_ylabel('Topics')

    for i, w in enumerate(weights):
        if w > 0.01:
            plt.text(i, w, str(i + 1), ha='center', va='bottom')



from collections import Counter
from .common import canonical


def generate_mapping(tokens, stopwords):
    stopwords = set(stopwords)
    mapping = dict()
    result = dict()

    for token, _count in Counter(tokens).most_common():
        key = canonical(token, aggresive=True, stopwords=stopwords)

        if key not in mapping:
            mapping[key] = token
        else:
            result[token] = mapping[key]

    return result



from time import sleep
from typing import Tuple, Optional
from urllib.parse import quote_plus
import logging
import requests
import shelve

from ..common import progress_bar
from ..types import Document, Author, DocumentSet, DocumentIdentifier


def extract_id(item):
    if item is None or not item.get('title'):
        return None

    return DocumentIdentifier(
            item['title'],
            doi=item.get('doi'),
            arxivid=item.get('arxivId'),
            s2id=item.get('paperId'),
    )


def extract_ids(items):
    if not items:
        return None

    return list(filter(None, map(extract_id, items)))


class ScholarAuthor(Author):
    def __init__(self, entry):
        self.entry = entry

    @property
    def name(self):
        return self.entry.get('name')

    @property
    def orcid(self):
        return None


class ScholarDocument(Document):
    def __init__(self, entry):
        super().__init__(extract_id(entry))
        self.entry = entry

    @property
    def title(self) -> str:
        return self.entry.get('title')

    @property
    def authors(self):
        authors = self.entry.get('authors')
        if not authors:
            return None

        return [ScholarAuthor(a) for a in authors if a]

    @property
    def publication_year(self):
        return self.entry.get('year')

    @property
    def publication_source(self):
        return self.entry.get('venue')

    @property
    def abstract(self):
        return self.entry.get('abstract')

    @property
    def citations(self):
        return extract_ids(self.entry.get('citations'))

    @property
    def citation_count(self):
        return self.entry.get('numCitedBy')

    @property
    def references(self):
        return extract_ids(self.entry.get('references'))

    def __repr__(self):
        return f'<{self.title}>'

    @staticmethod
    def load(id):
        return fetch_semanticscholar(id)


S2_PAPER_URL = 'http://api.semanticscholar.org/v1/paper/'
S2_QUERY_URL = 'https://api.semanticscholar.org/graph/v1/paper/search'
CACHE_FILE = '.semantischolar'
DEFAULT_TIMEOUT = 3.05  # 100 requests per 5 minutes


def request_results(query, offset, cache, timeout=DEFAULT_TIMEOUT):
    cache_key = f'results={query};{offset}'
    if cache_key in cache:
        return cache[cache_key]

    url = S2_QUERY_URL
    params = dict(offset=offset, query=query, limit=100)
    reply = requests.get(url, params=params)
    response = reply.json()

    if 'data' not in response:
        msg = response.get('error') or response.get('message') or 'unknown'
        raise Exception(f'error while fetching {reply.url}: {msg}')

    cache[cache_key] = response
    return response


def request_paper(key, cache, timeout=DEFAULT_TIMEOUT):
    cache_key = f'paper={key}'
    if cache_key in cache:
        return cache[cache_key]

    url = S2_PAPER_URL + quote_plus(key)

    try:
        sleep(timeout)
        data = requests.get(url).json()
    except Exception as e:
        logging.warn(f'failed to retreive {key}: {e}')
        return None

    if 'paperId' in data:
        cache[cache_key] = data
        return data
    else:
        msg = data.get('error') or data.get('message') or 'unknown error'
        logging.warn(f'failed to retreive {key}: {msg}')
        return None


def fetch_semanticscholar(key: set) -> Optional[Document]:
    """Fetch SemanticScholar metadata for the given key. The key can be
    one of the following (see `API reference
    <https://www.semanticscholar.org/product/api>`_):

    * DOI
    * S2 paper ID
    * ArXiv ID (example format: `arXiv:1705.10311`)
    * MAG ID (example format: `MAG:112218234`)
    * ACL ID (example format: `ACL:W12-3903`)
    * PubMed ID (example format: `PMID:19872477`)
    * Corpus ID (example format: `CorpusID:37220927`)

    :returns: The `Document` if it was found and `None` otherwise.
    """

    if key is None:
        return None

    with shelve.open(CACHE_FILE) as cache:
        if isinstance(key, DocumentIdentifier):
            data = None
            if data is None and key.s2id:
                data = request_paper(key.s2id, cache)

            if data is None and key.doi:
                data = request_paper(key.doi, cache)

            if data is None and key.pubmed:
                data = request_paper(f'PMID:{key.pubmed}', cache)

            if data is None and key.arxivid:
                data = request_paper(f'arXiv:{key.arxivid}', cache)
        else:
            data = request_paper(key, cache)

    if data is None:
        return None

    return ScholarDocument(data)


def refine_semanticscholar(docs: DocumentSet
                           ) -> Tuple[DocumentSet, DocumentSet]:
    """Attempt to fetch SemanticScholar metadata for each document in the
    given set based on their DOIs. Returns a tuple containing two sets: the
    documents available on SemanticScholar and the remaining documents that
    were not found or do not have a DOI.
    """
    def callback(doc):
        if isinstance(doc, ScholarDocument):
            return doc

        return fetch_semanticscholar(doc.id)

    return docs._refine_docs(callback)


def search_semanticscholar(query: str, *, limit: int = None) -> DocumentSet:
    """ Submit the given query to SemanticScholar and return the results
    as a `DocumentSet`.
    """

    if not query:
        raise Exception('invalid query: {query}')

    docs = []

    with shelve.open(CACHE_FILE) as cache:
        offset = 0
        paper_ids = []

        while True:
            data = request_results(query, offset, cache)
            if not data:
                break

            records = data['data']
            offset += len(records)

            for record in records:
                paper_ids.append(record['paperId'])

            if limit is not None and len(paper_ids) > limit:
                paper_ids = paper_ids[:limit]
                break

        for paper_id in progress_bar(paper_ids):
            doc = request_paper(paper_id, cache)

            if doc:
                docs.append(ScholarDocument(doc))
            else:
                logging.warn(f'could not find paper id {paper_id}')

    return DocumentSet(docs)



from ..types import Document, DocumentSet, DocumentIdentifier, Author
# from bibtexparser.customization import convert_to_unicode
from bibtexparser.latexenc import latex_to_unicode
import bibtexparser
from datetime import date
import re

MONTHS = dict(
        jan=1, january=1,
        feb=2, feburary=2,
        mar=3, march=3,
        apr=4, april=4,
        may=5,
        jun=6, june=6,
        jul=7, july=7,
        aug=8, august=8,
        sep=9, september=9,
        oct=10, october=10,
        nov=11, november=11,
        dec=12, december=12,
)


def find_doi(entry):
    def extract(val):
        if isinstance(val, str):
            # The following regex is recommended by crossref to detect DOIs
            # http://crossref.org/blog/dois-and-matching-regular-expressions/
            #   /^10.\d{4,9}/[-._;()/:A-Z0-9]+$/i
            pattern = '10[.][0-9]{4,9}/[-._;()/:a-zA-Z0-9]{5,}'

            if match := re.search(pattern, val):
                return match[0]

        return None

    if result := entry.get('doi', '').strip():
        doi = result
    else:
        doi = None

        for key in ['doi', 'link', 'url', 'howpublished']:
            if key in entry:
                if doi := extract(entry[key]):
                    break

        if not doi:
            return None

    # Remove URL prefix
    prefix = 'https://doi.org/'
    if doi.startswith(prefix):
        doi = doi[len(prefix):]

    return doi


class BibDocument(Document):
    def __init__(self, entry):
        title = entry['title']
        attr = dict(
                doi=find_doi(entry),
                pubmed=entry.get('pmid'),
        )

        super().__init__(DocumentIdentifier(title, **attr))
        self.entry = entry

    @property
    def key(self) -> str:
        return self.entry['ID']

    @property
    def title(self) -> str:
        return self.entry.get('title').strip('{}')

    @property
    def authors(self):
        content = self.entry.get('author')
        if not content:
            return None

        content = re.sub('[ \r\n\t]+', ' ', content)
        names = re.split(' (?:and|And|AND) ', content)

        if not names:
            return None

        # remove "other" from the list
        if names[-1] == 'others':
            names = names[:-1]

        return [BibAuthor(name) for name in names]

    @property
    def publisher(self):
        return self.entry.get('publisher')

    @property
    def language(self):
        return self.entry.get('language')

    @property
    def publication_date(self):
        year = self.publication_year
        month = self.publication_month or 1

        if year is None:
            return None

        return date(year, month, 1)

    @property
    def publication_year(self):
        try:
            year = int(self.entry['year'])
        except Exception:
            return None

        # Sometimes year is in 19xx format.
        if year < 100:
            year += 1900

        return year

    @property
    def publication_month(self):
        key = self.entry.get('month', '').lower()
        return MONTHS.get(key)  # name -> number

    @property
    def publication_source(self):
        if result := self.entry.get('journal'):
            return result

        if result := self.entry.get('booktitle'):
            return result

        return None

    @property
    def keywords(self):
        content = self.entry.get('keywords')
        if not content:
            return None

        return [w.strip().lower() for w in re.split('[;,\n\r]+', content) if w]

    @property
    def abstract(self):
        return self.entry.get('abstract')


class BibAuthor(Author):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return f'<{self._name}>'


def load_bibtex(path: str) -> DocumentSet:
    """Load the bibtex file at the given `path` as a `DocumentSet`.
    """
    def decode(entry):
        if isinstance(entry, list):
            return [decode(e) for e in entry]
        elif isinstance(entry, dict):
            return dict((k, decode(v)) for k, v in entry.items())
        else:
            # TODO: latex_to_unicode sometimes fails with exception. I do
            # not understand why, but let's just sweep it under the rug
            # for now ok? Great.
            try:
                return latex_to_unicode(entry)
            except Exception:
                return entry

    parser = bibtexparser.bparser.BibTexParser(common_strings=True)
    parser.customization = decode

    with open(path) as f:
        data = bibtexparser.load(f, parser=parser)

    docs = [BibDocument(e) for e in data.entries if e.get('title')]
    return DocumentSet(docs)



from .scopus import search_scopus, refine_scopus, fetch_scopus
from .bibtex import load_bibtex
from .semanticscholar import fetch_semanticscholar, search_semanticscholar, \
                             refine_semanticscholar
from .crossref import fetch_crossref, refine_crossref
from .ieee import load_ieee_csv
from .springer import load_springer_csv
from .dblp import search_dblp
from .ris import load_ris_file

__all__ = [
    'refine_crossref',
    'fetch_crossref',
    'refine_semanticscholar',
    'search_semanticscholar',
    'fetch_semanticscholar',
    'refine_scopus',
    'search_scopus',
    'fetch_scopus',
    'load_bibtex',
    'load_ieee_csv',
    'load_springer_csv',
    'load_ris_file',
    'search_dblp',
]



from ..common import progress_bar, canonical
from ..types import Document, DocumentSet, DocumentIdentifier, Author, \
                   Affiliation
from collections import defaultdict
from datetime import date
from typing import Tuple, Optional
import logging
import random
import shelve


SCOPUS_CACHE = '.scopus'


class ScopusAuthor(Author):
    def __init__(self, name, affiliations):
        self._name = name
        self._affiliations = affiliations

    @property
    def name(self):
        return self._name

    @property
    def affiliations(self):
        return self._affiliations


class ScopusAffiliation(Affiliation):
    def __init__(self, affiliation):
        self._affiliation = affiliation

    @property
    def name(self) -> str:
        return self._affiliation.organization

    @property
    def country(self):
        return self._affiliation.country or None

    def __repr__(self):
        return f'<{self.name}>'


class ScopusDocument(Document):
    @staticmethod
    def from_identifier(id, id_type, view='FULL'):
        from pybliometrics.scopus import AbstractRetrieval
        from pybliometrics.scopus.exception import Scopus404Error

        with shelve.open(SCOPUS_CACHE) as cache:
            key = id + '_found'
            if cache.get(key) is False:
                raise Scopus404Error()

            try:
                result = AbstractRetrieval(id, id_type=id_type, view=view)
                return ScopusDocument(result)
            except Scopus404Error:
                cache[key] = False
                raise

    @staticmethod
    def from_eid(eid, **kwargs):
        return ScopusDocument.from_identifier(eid, 'eid', **kwargs)

    @staticmethod
    def from_doi(doi, **kwargs):
        return ScopusDocument.from_identifier(doi, 'doi', **kwargs)

    def __init__(self, doc):
        identifier = DocumentIdentifier(
                doc.title,
                doi=doc.doi,
                pubmed=doc.pubmed_id,
                eid=doc.eid,
        )

        super().__init__(identifier)
        self.doc = doc

    @property
    def title(self):
        return self.doc.title or None

    @property
    def authors(self):
        if self.doc.authorgroup is not None:
            items = defaultdict(list)

            for aff in self.doc.authorgroup:
                name = f'{aff.indexed_name} (AUID: {aff.auid})'
                items[name].append(ScopusAffiliation(aff))

            return [ScopusAuthor(a, f) for a, f in items.items()]

        if self.doc.authors is not None:
            at = self.doc.authors
            return [ScopusAuthor(a.indexed_name, None) for a in at]

        return None

    @property
    def publisher(self):
        return self.doc.publisher or None

    @property
    def language(self):
        return self.doc.language or None

    @property
    def keywords(self):
        return self.doc.authkeywords or []

    @property
    def abstract(self):
        return self.doc.abstract or self.doc.description or None

    @property
    def citation_count(self):
        if self.doc.citedby_count is not None:
            return int(self.doc.citedby_count)
        return None

    @property
    def references(self):
        refs = []

        if not self.doc.references:
            return None

        for ref in self.doc.references:
            if ref.id:
                eid = '2-s2.0-' + ref.id  # id is Scopus id without prefix
            else:
                eid = None

            title = ref.title or ref.sourcetitle
            doi = ref.doi

            # Some references have no title, DOI, or Scopus id. This
            # makes them essentially useless.
            if title or doi or eid:
                refs.append(DocumentIdentifier(
                    title,
                    eid=eid,
                    doi=doi,
                ))

        return refs

    @property
    def publication_source(self):
        return self.doc.confname or self.doc.publicationName or None

    @property
    def source_type(self):
        return self.doc.aggregationType

    @property
    def publication_date(self):
        if self.doc.confdate:
            year, month, day = self.doc.confdate[0]
            return date(year, month, day)

        if self.doc.coverDate:
            try:
                year, month, day = self.doc.coverDate.split('-')
                return date(int(year), int(month), int(day))
            except Exception:
                pass

        return None

    def __repr__(self):
        return f'<{self.title}>'


def fetch_scopus(key: str) -> Optional[Document]:
    """ Fetch the document on Scopus for the given key. The key can be one of
    the following options:

    * DOI
    * Scopus EID or Scopus ID
    * PII
    * Pubmed-ID
    """
    return ScopusDocument.from_identifier(key, None)


def search_scopus(query: str, *, limit: int = None) -> DocumentSet:
    """ Submit the given query to the Scopus API.

    :param limit: Restrict results the first `limit` documents.
    """
    from pybliometrics.scopus import ScopusSearch

    search = ScopusSearch(query, view='STANDARD')
    eids = list(search.get_eids())
    docs = []

    if limit is not None and len(eids) > limit:
        random.seed(0)
        random.shuffle(eids)
        eids = eids[:limit]

    for eid in progress_bar(eids):
        doc = ScopusDocument.from_eid(eid)
        docs.append(doc)

    return DocumentSet(docs)


def refine_scopus(docs: DocumentSet, *, search_title=True
                  ) -> Tuple[DocumentSet, DocumentSet]:
    """Attempt to fetch Scopus metadata for each document in the given
    set. Returns a tuple containing two sets: the documents available on
    Scopus and the remaining documents not found on Scopus.


    Documents are retrieved based on their identifier (DOI, Pubmed ID, or
    Scopus ID). Documents without a unique identifier are retrieved by
    performing a fuzzy search based on their title. This is not ideal
    and can lead to false positives (i.e., another document is found having
    the same title), thus it can be disabled if necessary.

    :param search_title: Flag to toggle searching by title.
    """
    from pybliometrics.scopus import ScopusSearch

    def callback(doc):
        id = doc.id
        if isinstance(doc, ScopusDocument):
            return doc

        if doi := id.doi:
            try:
                return ScopusDocument.from_doi(doi)
            except Exception as e:
                logging.warn(f'no document found for DOI {doi}: {e}')
                return None

        title = canonical(id.title)
        if len(title) > 10 and search_title:
            query = f'TITLE({title})'
            response = ScopusSearch(query, view='STANDARD', download=False)
            nresults = response.get_results_size()

            if nresults > 0 and nresults < 10:
                response = ScopusSearch(query, view='STANDARD')

                for record in response.results or []:
                    if canonical(record.title) == title:
                        return ScopusDocument.from_eid(record.eid)

        return None

    return docs._refine_docs(callback)



import csv

from ..types import Document, DocumentSet, DocumentIdentifier


class SpringerDocument(Document):
    def __init__(self, entry):
        doi = entry['Item DOI'] or None
        title = entry['Item Title']

        super().__init__(DocumentIdentifier(
                title,
                doi=doi
        ))
        self.entry = entry

    @property
    def title(self) -> str:
        return self.entry['Item Title']

    @property
    def authors(self):
        # While Springer does provide an authors field, all names are
        # concatenated into one long string without seperators, making
        # it impossible to seperate them. Too bad.
        # authors = self.entry.get('Authors')
        return None

    @property
    def publisher(self):
        return 'springer'

    @property
    def publication_year(self):
        try:
            return int(self.entry['Publication Year'])
        except Exception:
            return None


def load_springer_csv(path: str) -> DocumentSet:
    """ Load CSV file exported from
    `Springer Link <https://link.springer.com/>`_.
    """
    with open(path, newline='') as f:
        lines = csv.DictReader(f)
        docs = [SpringerDocument(line) for line in lines]
        return DocumentSet(docs)



from __future__ import annotations
from abc import ABC, abstractmethod
from datetime import date
from typing import Optional, List
import numpy as np
import pandas as pd
import random
import re

from .common import fuzzy_match, canonical, progress_bar


class DocumentSet:
    """ Represents a set of documents.

    `DocumentSet` stores a list of `Document` objects. Optionally, a pandas
    data frame can be provided which stores additional properties on the
    documents.

    All set operations are accepted by `DocumentSet` (union, intersection,
    difference), allowing for new sets to be created from existing sets.

    Note that a `DocumentSet` is immutable and its content cannot be changed.
    Instead, most methods below return a new `DocumentSet` instead of
    performing modifications in-place.
    """

    def __init__(self, docs, data=None):
        """ Construct a new `DocumentSet`.

        :param docs: A list (or iterator) of `Document` objects.
        :param data: Additional metadata associated with the documents. This
                     can be either a `pandas.DataFrame` or something which
                     is accepted by the pandas DataFrame constructor.
        """
        docs = list(docs)

        if data is None:
            data = pd.DataFrame(index=range(len(docs)))
        else:
            data = pd.DataFrame(data)

        assert len(data) == len(docs)
        self.data = data
        self.docs = docs

    def _refine_docs(self, callback):
        new_indices = []
        new_docs = []
        old_indices = []
        old_docs = []

        for i, doc in enumerate(progress_bar(self.docs)):
            new_doc = callback(doc)

            if new_doc is not None:
                new_indices.append(i)
                new_docs.append(new_doc)
            else:
                old_indices.append(i)
                old_docs.append(doc)

        new_data = self.data.iloc[new_indices]
        old_data = self.data.iloc[old_indices]

        # FIX: not forget to reset the index
        new_data.reset_index(drop=True, inplace=True)
        old_data.reset_index(drop=True, inplace=True)

        return DocumentSet(new_docs, new_data), DocumentSet(old_docs, old_data)

    def add_property(self, name: str, values) -> DocumentSet:
        """ Returns a new set which has an additional property added.

        :param name: Name of the new property.
        :param values: List of values. Should be the same length as the
                       number of documents in this set.
        :returns: The new document set.
        """
        data = self.data.copy(deep=False)
        data[name] = values
        return DocumentSet(self.docs, data)

    def remove_property(self, name: str) -> DocumentSet:
        """ Returns a new set which has the given property removed.

        :param name: Name of the property.
        :returns: The new document set.
        """
        data = self.data.copy(deep=False)
        data.drop(name)
        return DocumentSet(self.docs, data)

    def filter_docs(self, predicate) -> DocumentSet:
        """ Returns a new set for which the provided predicate returned `True`.

        :param predicate: A function `Document -> bool`.
        :returns: The new document set.
        """
        return self.filter(lambda doc, _: predicate(doc))

    def filter(self, predicate) -> DocumentSet:
        """ Returns a new set for which the provided predicate returned `True`.

        :param predicate: A function `Document, dict -> bool`. The provided
                          dict stores the properties of the document.
        :returns: The new document set.
        """
        indices = []

        for doc, record in zip(self.docs, self.data.itertuples()):
            if predicate(doc, record):
                indices.append(record.Index)

        return self.select(indices)

    def select(self, indices) -> DocumentSet:
        """ Returns a new set which contains only the documents at the
        provided indices.

        :param indices: Any input accepted by `pandas.DataFrame.iloc` such
                        as a list of integer.
        :returns: The new document set.
        """

        data = self.data.iloc[indices]
        docs = [self.docs[i] for i in data.index]
        data = data.reset_index(drop=True)
        return DocumentSet(docs, data)

    def _intersect_indices(self, other):
        haystack = [d.id for d in other]
        left = []
        right = []

        for i, doc in enumerate(self):
            needle = doc.id

            for j, id in enumerate(haystack):
                if id.matches(needle):
                    left.append(i)
                    right.append(j)
                    break

        return left, right

    def _zip_with(self, left, other, right):
        assert len(left) == len(right)
        data = dict()

        for key, column in self.data.iloc[left].items():
            column = column.copy().reset_index(drop=True)
            data[key] = column

        for key, column in other.data.iloc[right].items():
            column = column.copy().reset_index(drop=True)

            if key in data:
                data[key] = np.maximum(data[key], column)
            else:
                data[key] = column

        return pd.DataFrame(index=range(len(left)), data=data)

    def intersect(self, other: DocumentSet) -> DocumentSet:
        """ Returns a new set which contains the documents provided in
        both `self` and `other`. This is also available as the `&` operator.

        :returns: The new document set.
        """
        if not self or not other:
            return DocumentSet([])

        left, right = DocumentSet._intersect_indices(self, other)

        docs = [self.docs[i] for i in left]  # Select docs from left?
        data = DocumentSet._zip_with(self, left, other, right)
        return DocumentSet(docs, data)

    def difference(self, other: DocumentSet) -> DocumentSet:
        """ Returns a new set which contains the documents provided in
        `self` but not in `other`. This is also available as the `-` operator.

        :returns: The new document set.
        """
        if not other or not self:
            return self

        indices, _ = DocumentSet._intersect_indices(self, other)
        return self.select(sorted(set(range(len(self))) - set(indices)))

    def union(self, other: DocumentSet) -> DocumentSet:
        """ Returns a new set which contains the documents provided in
        either `self` and `other`. Duplicate documents in `other` that also
        appear in `self` are discarded. This is also available as the `|`
        operator.

        :returns: The new document set.
        """
        if not other:
            return self
        if not self:
            return other

        left, right = DocumentSet._intersect_indices(self, other)
        if not left:
            return DocumentSet.concat(self, other)

        docs = [self.docs[i] for i in left]  # Select docs from left?
        data = DocumentSet._zip_with(self, left, other, right)
        middle = DocumentSet(docs, data)

        left = self.select(sorted(set(range(len(self))) - set(left)))
        right = other.select(sorted(set(range(len(other))) - set(right)))

        return DocumentSet.concat(middle, DocumentSet.concat(left, right))

    def concat(self, other: DocumentSet) -> DocumentSet:
        """ Returns a new set which does contain the documents provided in
        either `self` and `other`. Duplicate documents are not removed, see
        `union` instead. This is also available as the `+` operator.

        :returns: The new document set.
        """
        def default_val(dtype):
            c = dtype.char
            if c == '?':
                return False
            if c in 'bBiu':
                return 0
            if c in 'fc':
                return float('nan')
            if c in 'SaU':
                return ''

            return None

        if not self:
            return other

        if not other:
            return self

        left = self.data.copy()
        right = other.data.copy()

        for col in left:
            if col not in right:
                dtype = left[col].dtype
                right[col] = np.array(default_val(dtype)).astype(dtype)

        for col in right:
            if col not in left:
                dtype = right[col].dtype
                left[col] = np.array(default_val(dtype)).astype(dtype)

        data = pd.concat([left, right], ignore_index=True)
        docs = self.docs + other.docs
        return DocumentSet(docs, data)

    def unique(self) -> DocumentSet:
        """ Returns a new set which has all duplicate documents removed.

        :returns: The new document set.
        """
        indices = []

        for i, doc in enumerate(self.docs):
            found = False
            needle = doc.id

            for other in self[:i]:
                if other.id.matches(needle):
                    other._identifier = other._identifier.merge(needle)
                    found = True
                    break

            if not found:
                indices.append(i)

        return self.select(indices)

    def sample(self, n, seed=0) -> DocumentSet:
        """ Returns a new set which contains `n` randomly chosen documents
        from `self`.

        :returns: The new document set.
        """
        if len(self) <= n:
            return self

        random.seed(seed)
        indices = random.sample(len(self), n)
        indices.sort()
        return self.select(indices)

    def itertuples(self):
        """ Returns an iterator over `(Document, dict)` tuples, where the
        `dict` contains the properties of this document.
        """
        return zip(self.docs, self.data.itertuples())

    def __or__(self, other):
        """ Alias for `DocumentSet.union` """
        return self.union(other)

    def __and__(self, other):
        """ Alias for `DocumentSet.intersect` """
        return self.intersect(other)

    def __add__(self, other):
        """ Alias for `DocumentSet.concat` """
        return self.concat(other)

    def __sub__(self, other):
        """ Alias for `DocumentSet.difference` """
        return self.difference(other)

    def __len__(self):
        """ Returns the number of documents in this set """
        return len(self.docs)

    def __getitem__(self, key):
        """ Returns different things depending on the key type:

        * `str`: The property named `key` is returned.
        * `int`: The document at position `key` is returned.
        * otherwise: The call is forwarded to `DocumentSet.select`.
        """
        if isinstance(key, str):
            return self.data[key]
        elif np.issubdtype(type(key), np.integer):  # any type of integer works
            return self.docs[int(key)]
        else:
            return self.select(key)

    def __iter__(self):
        """ Returns an iterator over `Document` objects in this set. """
        return iter(self.docs)

    def __bool__(self):
        return bool(len(self))

    def __repr__(self):
        return f'<{len(self)} documents>'


class DocumentIdentifier:
    """ Represents an identifier for a document.

    Uniquely identifing an scientific document is often difficult since a
    single document might have multiple identifiers assigned to it (e.g., DOI,
    PubMed ID, Scopus ID, SemanticScholar ID) and not all data sources might
    provide all these identifiers. This class stores all possible identifiers
    that a document has.
    """

    def __init__(self, title, **attr):
        # Remove keys where value is None
        self._title = title
        self._attr = dict((k, v) for k, v in attr.items() if v)

    @property
    def title(self) -> Optional[str]:
        """ Returns the title. """
        return self._title

    @property
    def doi(self) -> Optional[str]:
        """ Returns the DOI (example: 10.1093/ajae/aaq063). """
        return self._attr.get('doi')

    @property
    def pubmed(self) -> Optional[str]:
        """ Returns the PubMed ID. """
        return self._attr.get('pubmed')

    @property
    def arxivid(self) -> Optional[str]:
        """ Returns the arXiv ID. """
        return self._attr.get('arxivid')

    @property
    def scopusid(self) -> Optional[str]:
        """ Returns the Scopus ID. """
        return self._attr.get('eid')

    @property
    def s2id(self) -> Optional[str]:
        """ Returns the Semantic Scholar ID. """
        return self._attr.get('s2id')

    def matches(self, other: DocumentIdentifier) -> bool:
        """ Returns `True` iff these two identifiers are equivalent


        Two documents are considered to be equivalent if all identifiers they
        have in common are equal. For example, if both documents have a DOI
        then these should be the same. If two documents have not a single
        identifier in common, a fuzzy match based on the title is performed.
        """
        n = 0

        # Two identifiers match if all keys that they have in common are equal
        for key in self._attr:
            if key in other._attr:
                if self._attr[key] != other._attr[key]:
                    return False
                n += 1

        if n > 0:
            return True

        # No identifiers in common
        return fuzzy_match(self._title, other._title)

    def merge(self, other) -> 'DocumentIdentifier':
        """ Returns a new `DocumentIdentifier` which adds the identifiers
        `others` to `self`.
        """
        attr = dict()
        attr.update(other._attr)
        attr.update(self._attr)
        return DocumentIdentifier(self._title, **attr)

    def __repr__(self):
        return f'<{self._title}, {self._attr}>'


class Document(ABC):
    """ Stores the metadata of a document.

    This is an interface which provides several methods which can be
    overridden by child classes. All methods can thus return `None`
    in case that method is not overridden.
    """
    def __init__(self, identifier: DocumentIdentifier):
        self._identifier = identifier

    @property
    def id(self) -> DocumentIdentifier:
        """ The `DocumentIdentifier` of this document. """
        return self._identifier

    @property
    @abstractmethod
    def title(self) -> str:
        """ The title of this document. """
        pass

    @property
    @abstractmethod
    def authors(self) -> Optional[List[Author]]:
        """ The authors of this document. """
        pass

    @property
    def affiliations(self) -> Optional[List[Affiliation]]:
        """ The affiliations associated with the authors of this document. """
        authors = self.authors

        if authors is not None:
            items = dict()

            for author in authors:
                for aff in author.affiliations:
                    items[aff.name] = aff

            return list(items.values())

        return None

    @property
    def publisher(self) -> Optional[str]:
        """ The publisher of this document. """
        return None

    @property
    def language(self) -> Optional[str]:
        """ The language this document is written in. """
        return None

    @property
    def publication_date(self) -> Optional[date]:
        """ The data of publication. """
        return None

    @property
    def publication_year(self) -> Optional[int]:
        """ The year of publication. """
        date = self.publication_date
        if date is None:
            return None
        return date.year

    @property
    def publication_source(self) -> Optional[str]:
        """ The name of the publication source (i.e., journal name,
        conference name, etc.)
        """
        return None

    @property
    def source_type(self) -> Optional[str]:
        """ The type of publication source (i.e., journal, conference
        proceedings, book, etc.)
        """
        return None

    @property
    def keywords(self) -> Optional[List[str]]:
        """ The keywords of this document. What exactly consistutes as
        keywords depends on the data source (author keywords, generated
        keywords, topic categories), but is should be a list of strings.
        """
        return None

    @property
    def abstract(self) -> Optional[str]:
        """ The abstract of this document. """
        return None

    @property
    def citation_count(self) -> Optional[int]:
        """ The number of citations that this document received. """
        return None

    @property
    def references(self) -> Optional[List[DocumentIdentifier]]:
        """ The list of other documents that are cited by this document. """
        return None

    @property
    def citations(self) -> Optional[List[DocumentIdentifier]]:
        """ The list of other documents that cite this document. """
        return None

    def mentions(self, term: str) -> bool:
        """ Returns `True` if this document mentions the given term in the
        title, abstract, or keywords.
        """
        pattern = r'(^|\s)' + re.escape(term) + r'($|\s)'
        flags = re.IGNORECASE
        keywords = self.keywords or []

        for text in [self.title, self.abstract] + keywords:
            if text and re.search(pattern, text, flags=flags):
                return True

        return False


class Affiliation(ABC):
    """ Represents the affiliation of an author """
    @property
    @abstractmethod
    def name(self) -> str:
        """ Name of the affiliation """
        pass

    @property
    def city(self) -> Optional[str]:
        """ City the affiliation is located in. """
        pass

    @property
    def country(self) -> Optional[str]:
        """ Country the affiliation is located in. """
        pass


class Author(ABC):
    """ Represents the author of a document. """
    @property
    @abstractmethod
    def name(self) -> str:
        """ The name of the author. """
        pass

    @property
    def orcid(self) -> Optional[str]:
        """ The ORCID of the author. """
        return None

    @property
    def s2id(self) -> Optional[str]:
        """ The SemanticScholar ID of the author. """
        return None

    @property
    def affiliations(self) -> 'Optional[list[Affiliation]]':
        """ The affiliations this author is associated with. """
        return None


class DocumentMapping:
    def __init__(self, docs=None):
        self.title = dict()
        self.doi = dict()
        self.eid = dict()

        if docs:
            for index, doc in enumerate(docs):
                self.add(doc.id, index)

    def add(self, doc: DocumentIdentifier, value):
        if doc.scopusid:
            self.eid[doc.scopusid] = value

        if doc.doi:
            self.doi[doc.doi] = value

        if doc.title:
            self.title[canonical(doc.title)] = value

    def get(self, doc: DocumentIdentifier):
        result = None

        if result is None and doc.scopusid:
            result = self.eid.get(doc.scopusid)

        if result is None and doc.doi:
            result = self.doi.get(doc.doi)

        if result is None and doc.title:
            result = self.title.get(canonical(doc.title))

        return result



import re
from unidecode import unidecode

try:
    from tqdm import tqdm

    def progress_bar(it):
        return tqdm(it)
except ImportError:
    def progress_bar(it):
        return it

STOPWORDS = set([
    '',
    'and',
    'at',
    'for',
    'in',
    'into',
    'of',
    'on',
    'onto',
    'over',
    'the',
    'to',
    'ltd',
    'corporation',
    'corp',
])


def canonical(key, aggresive=True, stopwords=None):
    if stopwords is None:
        stopwords = STOPWORDS

    if aggresive:
        key = unidecode(key).lower()

    tokens = re.split(r'[\W]+', key)
    new_tokens = []

    for token in tokens:
        if not token or token[0].isdigit():
            continue

        if aggresive and (token in stopwords or len(token) <= 1):
            continue

        new_tokens.append(token)

    return ' '.join(new_tokens)


def fuzzy_match(lhs, rhs):
    if lhs is None or rhs is None:
        return False

    return canonical(lhs) == canonical(rhs)


class FuzzyMatcher:
    def __init__(self, initial=None):
        mapping = dict()
        unmapping = dict()

        if initial is not None:
            for src, dst in initial.items():
                key = canonical(src)
                dst_key = canonical(dst)

                mapping[dst_key] = key
                mapping[key] = key
                unmapping[key] = dst

        self.mapping = mapping
        self.unmapping = unmapping

    def get(self, name):
        key = canonical(name)

        if key in self.mapping:
            return self.unmapping[self.mapping[key]]

        nice_name = canonical(name, False)
        self.mapping[key] = key
        self.unmapping[key] = nice_name
        return nice_name



from .common import FuzzyMatcher
from collections import defaultdict, OrderedDict
from .types import DocumentSet
import pandas as pd


def compute_histogram(docs, fun, keys=None, sort_by_key=False, groups=None,
                      limit=None):
    if isinstance(groups, list):
        groups = dict((v, v) for v in groups)

    if isinstance(groups, dict):
        data = dict()

        for name, value in groups.items():
            if isinstance(value, str):
                data[name] = docs.data.eval(value)
            else:
                values = list(value)
                assert len(values) == len(docs)
                data[name] = values

        groups = pd.DataFrame(data)
    elif groups is None:
        groups = pd.DataFrame(index=range(len(docs)))
    elif isinstance(groups, pd.DataFrame):
        pass

    assert len(groups) == len(docs)
    totals = defaultdict(lambda: 0)
    counts = defaultdict(lambda: 0)

    for doc, row in zip(docs, groups.itertuples()):
        for item in fun(doc):
            if item is None:
                continue

            totals[item] += 1

            for index, weight in enumerate(row[1:]):
                if weight:
                    counts[item, index] += weight

    if keys is None:
        totals = pd.Series(totals)
        totals = totals.sort_values(ascending=False)

        if limit is not None:
            totals = totals[:limit]

        if sort_by_key:
            totals = totals.sort_index()
    else:
        totals = OrderedDict((k, totals[k]) for k in keys)
        totals = pd.Series(totals)

    columns = groups.columns
    if len(columns):
        data = [[counts[item, index]
                for index in range(len(columns))]
                for item in totals.index]
    else:
        columns = ['Frequency']
        data = totals

    return pd.DataFrame(
            data,
            index=totals.index,
            columns=columns,
    )


def compute_groups_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    return compute_histogram(docs, lambda _: ['Frequency'], **kwargs).T


def compute_year_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    """ Compute a histogram of the number of documents published in each
    year. """
    years = [doc.publication_year for doc in docs]
    min_year = min(year for year in years if year)
    max_year = max(year for year in years if year)
    keys = list(range(min_year, max_year + 1))

    def extract(doc):
        yield doc.publication_year

    return compute_histogram(docs, extract, keys=keys, **kwargs)


def compute_number_authors_histogram(docs: DocumentSet, max_authors=10,
                                     **kwargs) -> pd.DataFrame:
    """ Compute a histogram of the number of authors per document.

    :param max_authors: If a document has more than `max_author` authors, it
                        is it is listed as a special "max authors" category.
    """
    keys = ['NA'] + list(range(1, max_authors + 1)) + [f'>{max_authors}']

    def extract(doc):
        n = len(doc.authors or [])
        if n == 0:
            yield 'NA'
        elif n > max_authors:
            yield '>{max_authors}'
        else:
            yield n

    return compute_histogram(docs, extract, keys=keys, **kwargs)


def compute_language_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    """ Compute a histogram of number of documents by language. """
    def extract(doc):
        if doc.language:
            yield doc.language

    return compute_histogram(docs, extract, **kwargs)


def default_mapper(mapper):
    if mapper is None:
        return FuzzyMatcher()
    elif type(mapper) is dict:
        return FuzzyMatcher(mapper)
    else:
        return mapper


def compute_source_histogram(docs: DocumentSet, mapper=None, **kwargs
                             ) -> pd.DataFrame:
    """ Compute a histogram of number of documents by publication source. """
    mapper = default_mapper(mapper)

    def extract(doc):
        source = doc.publication_source
        yield mapper.get(source) if source else '(unknown)'

    return compute_histogram(docs, extract, **kwargs)


def compute_source_type_histogram(docs: DocumentSet, **kwargs
                                  ) -> pd.DataFrame:
    """ Compute a histogram of number of documents by source type. """
    def extract(doc):
        yield doc.source_type or '(unknown)'

    return compute_histogram(docs, extract, **kwargs)


def compute_author_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    """ Compute a histogram of number of documents by author name. """
    def extract(doc):
        for author in doc.authors or []:
            yield author.name

    return compute_histogram(docs, extract, **kwargs)


def compute_author_affiliation_histogram(docs: DocumentSet, **kwargs
                                         ) -> pd.DataFrame:
    """ Compute a histogram of number of documents by (author name,
    affiliation name) combinations. This can help reduce conflicts where there
    are many authors of the same name working for different affiliations."""
    def extract(doc):
        for author in doc.authors or []:
            for affiliation in author.affiliations or []:
                if author.name and affiliation.name:
                    yield f'{author.name}, {affiliation.name}'

    return compute_histogram(docs, extract, **kwargs)


def compute_affiliation_histogram(docs: DocumentSet, mapper=None, **kwargs
                                  ) -> pd.DataFrame:
    """ Compute a histogram of number of documents by affiliation name. """
    mapper = default_mapper(mapper)

    def extract(doc):
        result = set()
        for author in doc.authors or []:
            for aff in author.affiliations or []:
                if aff.name:
                    result.add(mapper.get(aff.name))

        return result

    return compute_histogram(docs, extract, **kwargs)


def extract_country(aff):
    from .continent import COUNTRY_TO_CONTINENT

    # Sometimes affiliation has given country
    if country := aff.country:
        return country

    # Sometimes the country is in the affiliation name
    name = aff.name
    if name:
        for country in COUNTRY_TO_CONTINENT.keys():
            if country in name:
                return country

    return None


def compute_country_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    """ Compute a histogram of number of documents by affiliation country. """
    def extract(doc):
        result = set()
        for author in doc.authors or []:
            for aff in author.affiliations or []:
                if country := extract_country(aff):
                    result.add(country)

        return result

    return compute_histogram(docs, extract, **kwargs)


def compute_continent_histogram(docs: DocumentSet, **kwargs) -> pd.DataFrame:
    """ Compute a histogram of number of documents by affiliation
        continent.
    """
    from .continent import COUNTRY_TO_CONTINENT

    def extract(doc):
        result = set()
        for author in doc.authors or []:
            for aff in author.affiliations or []:
                if country := extract_country(aff):
                    country = country.strip().lower()

                    if country.startswith('the '):
                        country = country[4:]

                    if continent := COUNTRY_TO_CONTINENT.get(country):
                        result.add(continent)
                    else:
                        result.add('Other')

        return result

    return compute_histogram(docs, extract, **kwargs)



from .sources import *  # noqa: F403,F401
from .stats import \
        compute_year_histogram, \
        compute_author_histogram, \
        compute_author_affiliation_histogram, \
        compute_language_histogram, \
        compute_number_authors_histogram, \
        compute_source_histogram, \
        compute_source_type_histogram, \
        compute_affiliation_histogram, \
        compute_country_histogram, \
        compute_groups_histogram  # noqa: F401
from .plot import \
        plot_year_histogram, \
        plot_author_histogram, \
        plot_number_authors_histogram, \
        plot_author_affiliation_histogram, \
        plot_language_histogram, \
        plot_source_histogram, \
        plot_source_type_histogram, \
        plot_affiliation_histogram, \
        plot_country_histogram, \
        plot_continent_histogram, \
        plot_groups_histogram, \
        plot_word_distribution, \
        plot_topic_clouds, \
        plot_document_topics, \
        plot_embedding  # noqa: F401
from .network import \
        build_citation_network, \
        build_coauthor_network, \
        build_cocitation_network, \
        build_coupling_network, \
        plot_citation_network, \
        plot_coauthor_network, \
        plot_cocitation_network, \
        plot_coupling_network, \
        plot_network  # noqa: F401
from .nlp import \
        build_corpus, \
        train_nmf_model, \
        train_lda_model, \
        compute_word_distribution, \
        calculate_embedding  # noqa: F401

# __all__ = []



STOPWORDS = [
        "'d",
        "'ll",
        "'m",
        "'re",
        "'s",
        "'ve",
        'a',
        'about',
        'above',
        'achieve',
        'acm',
        'across',
        'after',
        'afterwards',
        'again',
        'against',
        'ain',
        'all',
        'almost',
        'alone',
        'along',
        'already',
        'also',
        'although',
        'always',
        'am',
        'among',
        'amongst',
        'amoungst',
        'amount',
        'an',
        'and',
        'another',
        'any',
        'anyhow',
        'anyone',
        'anything',
        'anyway',
        'anywhere',
        'appproach',
        'approach',
        'are',
        'aren',
        "aren't",
        'around',
        'as',
        'at',
        'available',
        'back',
        'based',
        'be',
        'became',
        'because',
        'become',
        'becomes',
        'becoming',
        'been',
        'before',
        'beforehand',
        'behind',
        'being',
        'below',
        'beside',
        'besides',
        'between',
        'beyond',
        'bill',
        'both',
        'bottom',
        'but',
        'by',
        'ca',
        'call',
        'can',
        'cannot',
        'cant',
        'co',
        'combination',
        'computer',
        'con',
        'copyright',
        'could',
        'couldn',
        "couldn't",
        'couldnt',
        'cry',
        'current',
        'd',
        'de',
        'describe',
        'detail',
        'did',
        'didn',
        "didn't",
        'discuss',
        'do',
        'does',
        'doesn',
        "doesn't",
        'doing',
        'don',
        "don't",
        'done',
        'down',
        'due',
        'during',
        'each',
        'eg',
        'eight',
        'either',
        'eleven',
        'else',
        'elsewhere',
        'empty',
        'enough',
        'etc',
        'even',
        'ever',
        'every',
        'everyone',
        'everything',
        'everywhere',
        'except',
        'few',
        'fifteen',
        'fifty',
        'fill',
        'find',
        'fire',
        'first',
        'five',
        'for',
        'former',
        'formerly',
        'forty',
        'found',
        'four',
        'from',
        'front',
        'full',
        'further',
        'get',
        'give',
        'go',
        'had',
        'hadn',
        "hadn't",
        'has',
        'hasn',
        "hasn't",
        'hasnt',
        'have',
        'haven',
        "haven't",
        'having',
        'he',
        'hence',
        'her',
        'here',
        'hereafter',
        'hereby',
        'herein',
        'hereupon',
        'hers',
        'herself',
        'high',
        'him',
        'himself',
        'his',
        'how',
        'however',
        'hpc',
        'hundred',
        'i',
        'ie',
        'ieee',
        'if',
        'impact',
        'in',
        'inc',
        'increase',
        'indeed',
        'interest',
        'into',
        'is',
        'isn',
        "isn't",
        'it',
        "it's",
        'its',
        'itself',
        'just',
        'keep',
        'key',
        'kg',
        'km',
        'large',
        'last',
        'latter',
        'latterly',
        'least',
        'less',
        'll',
        'ltd',
        'm',
        'ma',
        'made',
        'make',
        'many',
        'may',
        'me',
        'meanwhile',
        'method',
        'might',
        'mightn',
        "mightn't",
        'mill',
        'mine',
        'more',
        'moreover',
        'most',
        'mostly',
        'move',
        'much',
        'must',
        'mustn',
        "mustn't",
        'my',
        'myself',
        "n't",
        'name',
        'namely',
        'need',
        'needn',
        "needn't",
        'neither',
        'never',
        'nevertheless',
        'new',
        'next',
        'nine',
        'no',
        'nobody',
        'none',
        'noone',
        'nor',
        'not',
        'nothing',
        'now',
        'nowhere',
        'n‘t',
        'n’t',
        'o',
        'of',
        'off',
        'often',
        'on',
        'once',
        'one',
        'only',
        'onto',
        'or',
        'other',
        'others',
        'otherwise',
        'our',
        'ours',
        'ourselves',
        'out',
        'over',
        'own',
        'paper',
        'part',
        'per',
        'perhaps',
        'please',
        'present',
        'propose',
        'proposed',
        'provide',
        'put',
        'quite',
        'rather',
        're',
        'real',
        'really',
        'regarding',
        'results',
        's',
        'same',
        'say',
        'see',
        'seem',
        'seemed',
        'seeming',
        'seems',
        'serious',
        'set',
        'several',
        'shan',
        "shan't",
        'she',
        "she's",
        'should',
        "should've",
        'shouldn',
        "shouldn't",
        'show',
        'side',
        'significant',
        'since',
        'sincere',
        'six',
        'sixty',
        'size',
        'so',
        'some',
        'somehow',
        'someone',
        'something',
        'sometime',
        'sometimes',
        'somewhere',
        'springer',
        'still',
        'such',
        'system',
        't',
        'take',
        'ten',
        'than',
        'that',
        "that'll",
        'the',
        'their',
        'theirs',
        'them',
        'themselves',
        'then',
        'thence',
        'there',
        'thereafter',
        'thereby',
        'therefore',
        'therein',
        'thereupon',
        'these',
        'they',
        'thick',
        'thin',
        'third',
        'this',
        'those',
        'though',
        'three',
        'through',
        'throughout',
        'thru',
        'thus',
        'time',
        'to',
        'today',
        'together',
        'too',
        'top',
        'toward',
        'towards',
        'twelve',
        'twenty',
        'two',
        'un',
        'under',
        'unless',
        'until',
        'up',
        'upon',
        'us',
        'use',
        'used',
        'using',
        'various',
        've',
        'verlag',
        'very',
        'via',
        'was',
        'wasn',
        "wasn't",
        'we',
        'well',
        'were',
        'weren',
        "weren't",
        'what',
        'whatever',
        'when',
        'whence',
        'whenever',
        'where',
        'whereafter',
        'whereas',
        'whereby',
        'wherein',
        'whereupon',
        'wherever',
        'whether',
        'which',
        'while',
        'whither',
        'who',
        'whoever',
        'whole',
        'whom',
        'whose',
        'why',
        'will',
        'with',
        'within',
        'without',
        'won',
        "won't",
        'work',
        'would',
        'wouldn',
        "wouldn't",
        'y',
        'yet',
        'you',
        "you'd",
        "you'll",
        "you're",
        "you've",
        'your',
        'yours',
        'yourself',
        'yourselves',
        '‘d',
        '‘ll',
        '‘m',
        '‘re',
        '‘s',
        '‘ve',
        '’d',
        '’ll',
        '’m',
        '’re',
        '’s',
        '’ve',
]



from ..types import Document, Author, DocumentSet, DocumentIdentifier
import logging


def extract_title(attr):
    return attr.get('TI') or attr.get('T1') or attr.get('T2') or \
            attr.get('T3') or attr.get('TT')


class RISDocument(Document):
    def __init__(self, attr, keywords, authors):
        title = extract_title(attr)
        doi = attr.get('DO') or None

        super().__init__(DocumentIdentifier(title, doi=doi))
        self.attr = attr
        self.keyword_list = keywords
        self.author_list = authors

    @property
    def title(self) -> str:
        return extract_title(self.attr)

    @property
    def authors(self):
        return self.author_list

    @property
    def affiliations(self):
        return None

    @property
    def publisher(self):
        return self.attr.get('PB')

    @property
    def language(self):
        return self.attr.get('LA')

    @property
    def publication_year(self):
        try:
            return int(self.attr['PY'])
        except Exception:
            return None

    @property
    def publication_source(self):
        return None

    @property
    def keywords(self):
        return self.keyword_list

    @property
    def abstract(self):
        return self.attr.get('AB')


class RISAuthor(Author):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name


def load_ris_file(path: str) -> DocumentSet:
    """ Load the RIS file at the given `path` as a `DocumentSet`. """
    docs = []

    with open(path, newline='') as f:
        authors = []
        keywords = []
        attr = dict()

        for line in f:
            line = line.strip()
            if not line:  # ignore empty lines?
                continue

            if len(line) < 5 or line[2:5] != '  -':
                raise Exception(f'invalid RIS line: {line}')

            key = line[:2]
            value = line[5:].strip()

            if key == 'ER':
                docs.append(RISDocument(attr, keywords, authors))
                attr = dict()
                authors = []
                keywords = []
            elif key == 'KW':
                keywords.append(value)
            elif key in ['A1', 'A2', 'A3', 'A4', 'AU']:
                authors.append(RISAuthor(value))
            elif key in attr:
                logging.warn(f'Tag {key} appears multiple times ' +
                             f'("{value}" and "{attr[key]}")')
            else:
                attr[key] = value

    if attr:
        docs.append(RISDocument(attr, keywords, authors))

    return docs



from datetime import date
from time import sleep
from typing import Tuple, Optional
from urllib.parse import quote_plus
import logging
import re
import requests
import shelve

from ..types import Document, Author, DocumentSet, DocumentIdentifier, \
        Affiliation


class CrossRefAuthor(Author):
    def __init__(self, entry):
        self.entry = entry

    @property
    def name(self):
        given = self.entry.get('given', '')
        family = self.entry.get('family', '')

        if not given and not family:
            return None

        return f'{given} {family}'.strip()

    @property
    def orcid(self):
        return self.entry.get('ORCID')

    @property
    def affiliations(self):
        if entries := self.entry.get('affiliation'):
            return [CrossRefAffiliation(e) for e in entries]
        else:
            return None


class CrossRefAffiliation(Affiliation):
    def __init__(self, entry):
        self.entry = entry

    @property
    def name(self) -> str:
        return self.entry['name']


class CrossRefDocument(Document):
    def __init__(self, entry):
        title = entry.get('title')
        doi = entry.get('DOI')

        super().__init__(DocumentIdentifier(title, doi=doi))
        self.entry = entry

    @property
    def title(self) -> str:
        title = self.entry.get('title')
        if title:
            return re.sub(r'[\s]+', ' ', ' '.join(title))
        else:
            return None

    @property
    def authors(self):
        authors = self.entry.get('author', [])
        return [CrossRefAuthor(a) for a in authors]

    @property
    def publisher(self):
        return self.entry.get('publisher')

    @property
    def language(self):
        return self.get('language')

    @property
    def publication_date(self):
        try:
            parts = self.entry['published-print']['date-parts']
            year = int(parts[0])
            month = int(parts[1])
            return date(year, month, 1)
        except Exception:
            return None

    @property
    def publication_year(self):
        try:
            return int(self.entry['published-print']['date-parts'][0])
        except Exception:
            return None

    @property
    def publication_source(self):
        source = self.entry.get('container-title')
        if source:
            return source[0]
        else:
            return None

    @property
    def abstract(self):
        return self.entry.get('abstract')

    @property
    def citation_count(self):
        try:
            return int(self.entry['is-referenced-by-count'])
        except Exception:
            return None

    @property
    def references(self):
        output = []

        for ref in self.entry.get('reference', []):
            title = ref.get('unstructured')
            doi = ref.get('DOI')

            if title or doi:
                output.append(DocumentIdentifier(title, doi=doi))

        return output

    def __repr__(self):
        return f'<{self.title}>'

    @staticmethod
    def load(doi):
        return fetch_crossref(doi)


CACHE_FILE = '.crossref'
CROSSREF_URL = 'https://api.crossref.org/works/'


def fetch_crossref(doi: str, timeout=0.5) -> Optional[Document]:
    """Fetch the metadata for the given DOI from CrossRef.

    :returns: The `Document` or `None` if the DOI was not available.
    """
    def request():
        if not doi:
            return None

        with shelve.open(CACHE_FILE) as cache:
            if doi in cache:
                return cache[doi]

            url = CROSSREF_URL + quote_plus(doi)

            try:
                response = requests.get(url)
            except Exception as e:
                logging.warn(f'failed to retrieve {doi}: {e}')
                return None

            sleep(timeout)

            code = response.status_code
            if code == 200:
                try:
                    data = response.json()['message']
                except Exception as e:
                    logging.warn(f'invalid output from {url}: {e}')
                    return None
            elif code == 404:
                logging.warn(f'failed to retrieve {doi}: resource not found')
                data = None
            else:
                logging.warn(
                        f'failed to retrieve {doi} ({code}): {response.text}')
                return None

            cache[doi] = data
            return data

    data = request()
    return CrossRefDocument(data) if data else None


def refine_crossref(docs: DocumentSet, timeout=0.5
                    ) -> Tuple[DocumentSet, DocumentSet]:
    """Attempts to fetch metadata from CrossRef for each document in the given
    set. Returns a tuple of two sets: the documents retrieved from CrossRef
    and the remaining documents (i.e., without DOI or not found).

    :param timeout: Timeout in seconds between each request to throttle
        server communication.
    """
    def callback(doc):
        if isinstance(doc, CrossRefDocument):
            return doc

        return fetch_crossref(doc.id.doi, timeout)

    return docs._refine_docs(callback)



from ..types import Document, Author, DocumentSet, DocumentIdentifier, \
                    Affiliation
import csv


class IEEEDocument(Document):
    def __init__(self, entry):
        doi = entry['DOI'] or None
        title = entry['Document Title']

        super().__init__(DocumentIdentifier(
                title,
                doi=doi
        ))
        self.entry = entry

    @property
    def title(self) -> str:
        return self.entry.get('Document Title')

    @property
    def authors(self):
        authors = self.entry.get('Authors', '').split('; ')
        affs = self.entry.get('Author Affiliations', '').split('; ')
        assert len(authors) == len(affs)
        return [IEEEAuthor(a, b) for a, b in zip(authors, affs)]

    @property
    def publisher(self):
        return self.entry.get('Publisher')

    @property
    def publication_year(self):
        try:
            return int(self.entry['Publication Year'])
        except Exception:
            return None

    @property
    def keywords(self):
        # There are several sources of keywords. Should we only include
        # author keywords maybe? Or all possible keywords?
        keys = [
            'Author Keywords',
            'IEEE Terms',
            'INSPEC Controlled Terms',
            'INSPEC Non-Controlled Terms',
            'Mesh_Terms',
        ]

        keywords = []
        for key in keys:
            for word in self.entry.get(key, '').split(';'):
                if word:
                    keywords.append(word)

        return keywords

    @property
    def abstract(self):
        return self.entry.get('Abstract') or None

    @property
    def citation_count(self):
        try:
            return int(self.entry['Article Citation Count'])
        except Exception:
            return None


class IEEEAffiliation(Affiliation):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name


class IEEEAuthor(Author):
    def __init__(self, name, affiliation):
        self._name = name
        self._affiliation = affiliation

    @property
    def name(self):
        return self._name

    @property
    def affiliations(self):
        # Special case where affiliation is NA (not applicable)
        if self._affiliation == 'NA':
            return None

        return [IEEEAffiliation(self._affiliation)]


def load_ieee_csv(path: str) -> DocumentSet:
    """ Import CSV file exported from
    `IEEE Xplore <https://ieeexplore.ieee.org/search/searchresult.jsp>`_.
    """
    with open(path, newline='') as f:
        lines = csv.DictReader(f)
        docs = [IEEEDocument(line) for line in lines]
        return DocumentSet(docs)



from collections import defaultdict
from gensim.matutils import corpus2dense
from typing import List
import gensim
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import wordcloud

from .stopwords import STOPWORDS
from .types import DocumentSet


def filter_tokens(texts, predicate):
    for text in texts:
        yield [token for token in text if predicate(token)]


def preprocess_remove_short(texts, min_length=3):
    yield from filter_tokens(texts, lambda token: len(token) >= min_length)


def preprocess_remove_words(texts, stopwords):
    stopwords = set(w.strip() for w in stopwords)
    yield from filter_tokens(texts, lambda token: token not in STOPWORDS)


def preprocess_stopwords(texts):
    yield from preprocess_remove_words(texts, STOPWORDS)


def preprocess_replace_words(texts, replace_words):
    for text in texts:
        yield [replace_words.get(token, token) for token in text]


def preprocess_merge_bigrams(texts, bigrams):
    for text in texts:
        prev = None
        new_text = []

        for current in text:
            replacement = bigrams.get((prev, current))

            if replacement is not None:
                new_text.append(replacement)
                prev = None
            else:
                if prev is not None:
                    new_text.append(prev)
                prev = current

        yield new_text


def preprocess_merge_ngrams(texts, threshold):
    texts = list(texts)
    phrases = gensim.models.phrases.Phrases(texts, threshold=threshold,
                                            scoring='npmi')

    for text in texts:
        for word, score in phrases.analyze_sentence(text):
            if score is not None:
                text.append(word)

        yield text


def preprocess_outliers(texts, min_docs, max_docs):
    texts = list(texts)
    count = defaultdict(int)

    for text in texts:
        for token in set(text):
            count[token] += 1

    unwanted = set()
    for token, num_docs in count.items():
        if num_docs < min_docs or num_docs > max_docs:
            unwanted.add(token)

    yield from preprocess_remove_words(texts, unwanted)


def preprocess_smart_stemming(texts):
    texts = list(texts)
    stemmer = gensim.parsing.PorterStemmer()
    count = defaultdict(int)

    for text in texts:
        for token in text:
            count[token] += 1

    sorted_count = sorted(count.items(), key=lambda p: p[1], reverse=True)
    unstemming = dict()
    mapping = dict()

    for token, _ in sorted_count:
        stem = stemmer.stem(token)
        if stem in unstemming:
            mapping[token] = unstemming[stem]
        else:
            unstemming[stem] = token
            mapping[token] = token

    for text in texts:
        yield [mapping[token] for token in text]


class Corpus:
    """ Contains the word-frequency vectors for a set of documents. See
        `build_corpus` for more information.
    """
    def __init__(self, docs, filters, max_tokens):
        corpus = []

        for doc in docs:
            text = (doc.title or '') + ' ' + (doc.abstract or ' ')
            tokens = gensim.utils.tokenize(
                    text,
                    lowercase=True,
                    deacc=True)

            corpus.append(list(tokens))

        for f in filters:
            corpus = f(corpus)

        corpus = list(corpus)
        dic = gensim.corpora.Dictionary(corpus)
        dic.filter_extremes(keep_n=max_tokens)

        self.dictionary = dic
        """ The dictionary that maps indices to words
            (`gensim.corpora.Dictionary`).
        """

        self.frequencies = [dic.doc2bow(x) for x in corpus]
        """ List of word frequency vectors. Each vector corresponds to one
        document and consists of `(word_index, frequency)` tuples.
        """


def build_corpus(docs: DocumentSet, *, remove_words=None, min_word_length=3,
                 min_docs=5, max_docs_ratio=0.75, max_tokens=5000,
                 replace_words=None, custom_bigrams=None, ngram_threshold=None
                 ) -> Corpus:
    """ Build a `Corpus` object.

    This function takes the words from the title/abstract of the given
    documents, preprocesses the tokens, and returns a corpus consisting of a
    word frequency vector for each document. This preprocessing stage is
    highly customizable, thus it is advised to experiment with the many
    parameters.

    :param remove_words: list of words that should be ignored while building
                         the word frequency vectors.
    :param min_word_length: Words shorter than this are ignored.
    :param min_docs: Words that occur in fewer than this many documents are
                     ignored.
    :param max_docs_ratio: Words that occur in more than this document are
                           ignored. Should be ratio between 0 and 1.
    :param max_tokens: Only the top most common tokens are preserved.
    :param replace_words: Replace words by other words. Must be a `dict`
                          containing *original word* to *replacement word*
                          pairs.
    :param custom_bigrams: Add custom bigrams. Must be a `dict` where keys
                           are `(first, second)` tuples and values are
                           replacements. For example, the key can be
                           `("Big", "Data")` and the value `"BigData"`.
    :param ngram_threshold: Threshold used for n-gram detection. Is passed
                            to `gensim.models.phrases.Phrases` to detect
                            common n-grams.
    :returns: a `Corpus object`.
    """

    filters = []
    if custom_bigrams:
        filters.append(lambda w: preprocess_merge_bigrams(w, custom_bigrams))

    if remove_words:
        filters.append(lambda w: preprocess_remove_words(w, remove_words))

    if replace_words:
        filters.append(lambda w: preprocess_replace_words(w, replace_words))

    if min_word_length:
        filters.append(lambda w: preprocess_remove_short(w,
                       min_length=min_word_length))

    filters.append(preprocess_stopwords)

    if ngram_threshold is not None:
        filters.append(lambda w: preprocess_merge_ngrams(w, ngram_threshold))

    filters.append(preprocess_smart_stemming)

    if min_docs > 1 or max_docs_ratio < 1.0:
        max_docs = int(len(docs) * max_docs_ratio)
        filters.append(lambda w: preprocess_outliers(w, min_docs, max_docs))

    return Corpus(docs, filters, max_tokens)


class TopicModel:
    """ Topic model trained by one of the `train_*_model` functions. """

    def __init__(self, dictionary, doc2topic, topic2token):
        self.dictionary = dictionary

        self.doc2topic = doc2topic
        """ `N x T` matrix that stores the weights towards each of the T
        topics for the N documents.
        """

        self.topic2token = topic2token
        """ `T x M` matrix that stores the weights towards each of the M
        tokens for each of the T topics
        """

        self.num_topics = len(topic2token)

    def best_documents_for_topic(self, topic_id: int, limit=5) -> List[int]:
        """ Returns the documents that most strongly belong to the given
        topic.
        """
        return np.argsort(self.doc2topic[:, topic_id])[::-1][:limit]

    def document_topics(self, doc_id: int):
        """ Returns a numpy array indicating the weights towards the different
        topic for the given document. These weight sum up to one.
        """
        return self.doc2topic[doc_id]

    def best_token_weights_for_topic(self, topic_id: int, limit=5):
        """ Returns a list of `(token, weight)` tuples for the tokens that
        most strongly belong to the given topic.
        """
        dic = self.dictionary
        weights = self.topic2token[topic_id]

        indices = np.argsort(self.topic2token[topic_id])[::-1][:limit]
        return [(dic[i], weights[i]) for i in indices]

    def best_tokens_for_topic(self, topic_id: int, limit=5):
        """ Returns the top tokens that most strongly belong to the given
        topic. """
        results = self.best_token_weights_for_topic(topic_id, limit=limit)
        return [w for w, _ in results]

    def best_token_for_topic(self, topic_id: int) -> str:
        """ Returns the token that most strongly belongs to the given
        topic. """
        return self.best_tokens_for_topic(topic_id, limit=1)[0]

    def best_topic_for_token(self, token) -> int:
        """ Returns the topic index that most strongly belongs to the given
        token. """
        index = self.dictionary.token2id[token]
        return np.argmax(self.topic2token[:, index])

    def best_topic_for_documents(self) -> List[int]:
        """ Returns the topic for each document that most strongly belongs
        to that document.
        """
        return np.argmax(self.doc2topic, axis=1)


def train_nmf_model(corpus: Corpus, num_topics: int, seed=0, max_iter=500
                    ) -> TopicModel:
    """ Train a topic model using NMF.

    :param num_topics: The number of topics to train.
    :param seed: The seed used for random number generation.
    :param max_iter: The maximum number of iterations to use for training.
                     More iterations mean better results, but longer training
                     times.
    """
    import gensim.models.nmf

    dic = corpus.dictionary
    freqs = corpus.frequencies

    tfidf = gensim.models.tfidfmodel.TfidfModel(dictionary=dic)
    model = gensim.models.nmf.Nmf(
            list(tfidf[freqs]),
            num_topics=num_topics,
            passes=max_iter,
            random_state=seed,
            w_stop_condition=1e-9,
            h_stop_condition=1e-9,
            w_max_iter=50,
            h_max_iter=50)

    doc2topic = corpus2dense(model[freqs], num_topics).T
    topic2token = model.get_topics()

    return TopicModel(dic, doc2topic, topic2token)


def train_lda_model(corpus: Corpus, num_topics, seed=0, **kwargs
                    ) -> TopicModel:
    """ Train a topic model using LDA.

    :param num_topics: The number of topics to train.
    :param seed: The seed used for random number generation.
    :param kwargs: Arguments passed to `gensim.models.lda.LdaModel`.
    """
    from gensim.models.lda import LdaModel

    dic = corpus.dictionary
    freqs = corpus.frequencies

    model = LdaModel(list(corpus), **kwargs)

    doc2topic = corpus2dense(model[freqs], num_topics)
    topic2token = model.get_topics()

    return TopicModel(dic, doc2topic, topic2token)


def compute_word_distribution(corpus: Corpus, *, limit=None) -> pd.DataFrame:
    """ Returns dataframe that indicates, for each word, the number of
    documents that mention that word.
    """
    counter = defaultdict(int)
    dic = corpus.dictionary

    for vector in corpus.frequencies:
        for i, _ in vector:
            counter[i] += 1

    keys = sorted(counter, key=lambda k: counter[k], reverse=True)[:limit]
    if limit is not None:
        keys = keys[:limit]

    return pd.DataFrame(
            index=[dic[i] for i in keys],
            data=dict(count=[counter[i] for i in keys])
    )


def generate_topic_cloud(model: TopicModel, topic_id: int, cmap=None,
                         max_font_size=75, background_color='white'
                         ) -> wordcloud.WordCloud:
    """ Generate a word cloud for the given topic from the given topic model.

    :param cmap: The color map used to color the words.
    :param max_font_size: Size of the word which most strongly belongs to the
                          topic. The other words are scaled accordingly.
    :param background_color: Background color.
    """
    if cmap is None:
        cmap = 'Blues'

    cmap = plt.get_cmap(cmap)

    dic = model.dictionary
    vec = model.topic2token[topic_id]
    maximum = np.amax(vec)
    words = dict()

    for i in np.argsort(vec)[-100:]:
        if vec[i] > 0:
            words[dic[i]] = vec[i] / maximum

    def get_color(word, **kwargs):
        weight = kwargs['font_size'] / 75.0 * 0.7 + 0.3
        r, g, b = np.array(cmap(weight)[:3]) * 255
        return 'rgb({}, {}, {})'.format(int(r), int(g), int(b))

    wc = wordcloud.WordCloud(
            prefer_horizontal=True,
            max_font_size=max_font_size,
            background_color=background_color,
            color_func=get_color,
            scale=2,
            relative_scaling=0.5)

    wc.fit_words(words)

    return wc


def calculate_embedding(corpus: Corpus, *, rank=2, svd_dims=50, perplexity=30,
                        seed=0):
    """ Calculate a document embedding that assigns each document in the
    corpus a N-d position based on the word usage.

    :returns: A list of N-d tuples for the documents in the corpus.
    """
    from gensim.models.tfidfmodel import TfidfModel
    from sklearn.decomposition import TruncatedSVD
    from sklearn.manifold import TSNE

    dic = corpus.dictionary
    freqs = corpus.frequencies
    tfidf = corpus2dense(TfidfModel(dictionary=dic)[freqs], len(dic)).T

    if svd_dims is not None:
        svd = TruncatedSVD(n_components=svd_dims, random_state=seed)
        components = svd.fit_transform(tfidf)
    else:
        components = tfidf

    model = TSNE(rank, metric='cosine', square_distances=True,
                 perplexity=perplexity, random_state=seed)
    return model.fit_transform(components)
